from .ntypes import (
    t_delete,
    t_performance,
    t_storage,
    t_job,
    t_job_not_found,
    t_env,
    t_skip,
    t_send,
    t_error,
    error_busy,
    error_reject,
)
from IPython.core.display import display, HTML
from .runner.proxy import encode_proxy_url
from .runner.env_var import n_env
import pandas as pd
import traceback
import ipykernel
import subprocess
import requests
import base64
import copy
import os


enterprise_gateway = False
try:
    import enterprise_gateway.services.kernels.remotemanager  # noqa: F401

    enterprise_gateway = True  # noqa: F811
except ImportError:
    pass


class Manager:
    __filetype = None
    headers = {"Authorization": f"token {n_env.token}"}

    def __init__(self, filetype=None):
        self.headers = {"Authorization": f"token {n_env.token}"}
        self.__filetype = filetype
        self.set_runner_mode()

    def get_size(self):
        response = requests.get(
            f"{n_env.api}/{t_performance}/{t_storage}", headers=self.headers
        )
        data = response.json()
        if data and data.get(f"{t_storage}"):
            print("📝 Memory used", data.get(f"{t_storage}"))
        else:
            print("😢 Cannot get Memory usage", data.get(t_error))

    def reload_jobs(self):
        response = requests.put(
            f"{n_env.api}/{t_job}", headers=self.headers, params={"reload_jobs": "yes"}
        )
        data = response.json()
        if data and data.get("status") == t_send:
            print("✅ Jobs reloaded from save")
        else:
            print("😢 Jobs cannot be reloaded", data.get(t_error))

    def move_job(self, old_path, new_path):
        if self.is_production():
            print("No move_job done you are in production\n")
            return
        old_path = os.path.abspath(os.path.join(os.getcwd(), old_path))
        new_path = os.path.abspath(os.path.join(os.getcwd(), new_path))
        response = requests.put(
            f"{n_env.api}/{t_job}",
            headers=self.headers,
            params={
                "move": "yes",
                "type": self.__filetype,
                "old_path": old_path,
                "new_path": new_path,
            },
        )
        res = response.json()
        if res and res.get("status") == t_send:
            for ff in res.get("data"):
                print(f"✅ Job {ff.get('from')} moved to {ff.get('to')}\n")
        else:
            print(f"😢 Job cannot be moved to {new_path}", res.get(t_error))

    def is_production(self):
        return True if n_env.current.get("env") == "RUNNER" else False

    def set_runner_mode(self):
        try:
            r = requests.get(f"{n_env.api}/{t_env}", headers=self.headers, timeout=2)
            r.raise_for_status()
        except Exception:
            n_env.remote_mode = not n_env.remote_mode

    def get(self):
        public_url = f"{encode_proxy_url()}"
        print("You can check your current tasks list here :")
        display(HTML(f'<a href="{public_url}"">Manager</a>'))

    def get_logs(self):
        req = requests.get(url=f"{n_env.api}/logs", headers=self.headers)
        req.raise_for_status()
        jsn = req.json()
        return jsn

    def get_naas(self):
        naas_data = []
        try:
            r = requests.get(f"{n_env.api}/{t_job}", headers=self.headers)
            r.raise_for_status()
            naas_data = r.json()
        except requests.exceptions.ConnectionError:
            print(error_busy)
        except requests.exceptions.HTTPError as e:
            print(error_reject, e)
        return naas_data

    def get_value(self, path, do_print=True):
        try:
            r = requests.get(
                f"{n_env.api}/{t_job}",
                params={
                    "path": path,
                    "type": self.__filetype,
                    "light": True,
                },
                headers=self.headers,
            )
            r.raise_for_status()
            data = r.json()
            return data.get("value")
        except requests.exceptions.ConnectionError as err:
            if do_print:
                print(error_busy, err)
            raise
        except requests.exceptions.HTTPError as err:
            if do_print:
                print(error_reject, err)
            raise

    def notebook_path(self):
        if self.is_production():
            return os.path.join(n_env.server_root, n_env.current.get("path"))
        try:
            notebooks = self.running_notebooks()
            try:
                connection_file = os.path.basename(ipykernel.get_connection_file())
                kernel_id = connection_file.split("-", 1)[1].split(".")[0]
                if enterprise_gateway:
                    kernel_id = connection_file.split("-", 1)[1].split("_")[0]
                for notebook in notebooks:
                    if kernel_id in notebook["kernel_id"]:
                        return os.path.join(n_env.server_root, notebook["path"])
            except Exception:
                process_id = os.getpid()
                for notebook in notebooks:
                    if process_id in notebook["process_ids"]:
                        return os.path.join(n_env.server_root, notebook["path"])
        except Exception as e:
            tb = traceback.format_exc()
            print("notebook_path", e, tb)
        return None

    def __get_process_ids(self, name):
        try:
            child = subprocess.Popen(
                ["pgrep", "-f", name], stdout=subprocess.PIPE, shell=False
            )
            response = child.communicate()[0]
            return [int(pid) for pid in response.split()]
        except Exception:
            return []

    def running_notebooks(self):
        try:
            base_url = f"{n_env.user_url}/api/sessions"
            req = requests.get(url=base_url, headers=self.headers)
            req.raise_for_status()
            sessions = req.json()
            sessions = filter(lambda item: "notebook" in item["type"], sessions)
            notebooks = [
                {
                    "kernel_id": notebook["kernel"]["id"],
                    "path": notebook["notebook"]["path"],
                    "process_ids": self.__get_process_ids(notebook["kernel"]["id"]),
                }
                for notebook in sessions
            ]
            return notebooks
        except requests.exceptions.ConnectionError as e:
            print(e)
        except requests.exceptions.HTTPError as e:
            print(e)
        except Exception as e:
            tb = traceback.format_exc()
            print("running_notebooks", e, tb)
        return []

    def get_path(self, path):
        if path is not None:
            base_path = os.getcwd()
            if self.is_production():
                base_path = base_path.replace(n_env.path_naas_folder, "")
            return os.path.abspath(os.path.join(base_path, path))
        else:
            return self.notebook_path()

    def proxy_url(self, endpoint, token=None):
        public_url = encode_proxy_url(endpoint)
        if token:
            public_url = f"{public_url}/{token}"
        return public_url

    def __open_file(self, path):
        filename = os.path.basename(path)
        if not os.path.exists(path):
            raise FileNotFoundError(f"file doesn't exist {path}")
        data = open(path, "rb").read()
        encoded = base64.b64encode(data)
        return {"filename": filename, "data": encoded.decode("ascii")}

    def safe_filepath(self, path):
        path = os.path.join(n_env.server_root, path)
        return path

    def __save_file(self, path, filedata=None):
        dirname = os.path.dirname(path)
        filename = filedata["filename"] or os.path.basename(path)
        new_path = os.path.join(dirname, filename)
        if not filedata:
            return
        if os.path.exists(new_path):
            raise FileNotFoundError(f"file already exist {new_path}")
        f = open(new_path, "wb")
        decoded = base64.b64decode(filedata["data"])
        f.write(decoded)
        f.close()
        return new_path

    def clear_file(self, path=None, mode=None, histo=None):
        if not path and self.is_production():
            print("No clear_prod done you are in production\n")
            return
        prod_path = self.get_path(path)
        try:
            r = requests.delete(
                f"{n_env.api}/{t_job}",
                headers=self.headers,
                params={
                    "path": prod_path,
                    "type": self.__filetype,
                    "histo": histo,
                    "mode": mode,
                },
            )
            r.raise_for_status()
            res = r.json()
            if res.get("status") == t_error or res.get("status") == t_skip:
                raise ValueError(f"❌ Cannot clean your file {path}")
            for ff in res.get("data"):
                print(f"🕣 Your file {ff} has been remove from production.\n")
            return pd.DataFrame(data=res.get("files", []))
        except requests.exceptions.ConnectionError as err:
            print(error_busy, err)
            raise
        except requests.exceptions.HTTPError as err:
            print(error_reject, err)
            raise

    def list_prod(self, mode, path=None, display=True):
        if not path and self.is_production():
            print("No list_prod done you are in production\n")
            return []
        if not path:
            result = pd.DataFrame()
            prod_files = self.get_naas()
            for file in prod_files:
                home = os.environ["HOME"]
                file = os.path.join(home, file["path"])
                if isinstance(file, list):
                    file = file[-1]
                new_prod = self.list_prod("list_history", file, False)
                if isinstance(new_prod, pd.DataFrame):
                    dfs = [result, new_prod]
                    result = pd.concat(dfs)
            return result
        current_file = self.get_path(path)
        try:
            r = requests.get(
                f"{n_env.api}/{t_job}",
                headers=self.headers,
                params={"path": current_file, "type": self.__filetype, "mode": mode},
            )
            if r.status_code != 404:
                r.raise_for_status()
            res = r.json()
            if (
                res.get("status") == t_error or res.get("status") == t_skip
            ) and res.get("error") != t_job_not_found:
                raise ValueError(f"❌ Cannot list your file {path}")
            if res.get("files", None) and len(res.get("files", [])) > 0:
                return pd.DataFrame(data=res.get("files", []))
            else:
                if display:
                    print("No files found in prod")
                return []
        except requests.exceptions.ConnectionError as err:
            print(error_busy, err)
            raise
        except requests.exceptions.HTTPError as err:
            print(error_reject, err)
            raise

    def get_file(self, path=None, mode=None, histo=None):
        if self.is_production():
            print("No get_prod done you are in production\n")
            return
        current_file = self.get_path(path)
        filename = os.path.basename(current_file)
        try:
            r = requests.get(
                f"{n_env.api}/{t_job}",
                headers=self.headers,
                params={
                    "path": current_file,
                    "type": self.__filetype,
                    "mode": mode,
                    "histo": histo,
                },
            )
            r.raise_for_status()
            res = r.json()
            if res.get("status") == t_error or res.get("status") == t_skip:
                raise ValueError(f"❌ Cannot get your file {path}")
            new_path = self.__save_file(
                self.safe_filepath(current_file), res.get("file")
            )
            print(
                f"🕣 Your Notebook {mode or ''} {filename}, has been copied into your local folder.\n"
            )
            return new_path
        except requests.exceptions.ConnectionError as err:
            print(error_busy, err)
            raise
        except requests.exceptions.HTTPError as err:
            print(error_reject, err)
            raise

    def path(self, filetype):
        def mode_path(self, path):
            nonlocal filetype
            if self.is_production():
                # filename = os.path.basename(path)
                # dirname = os.path.dirname(path)
                # filename = f"{filetype}_{filename}"
                # type_path = os.path.join(dirname, filename)
                type_path = path
                return self.get_path(type_path)
            else:
                return path

        return mode_path

    def add_prod(self, obj, debug):
        if "type" in obj and "path" in obj and "params" in obj and "value" in obj:
            new_obj = copy.copy(obj)
            dev_path = obj.get("path")
            new_obj["path"] = self.get_path(dev_path)
            new_obj["file"] = self.__open_file(dev_path)
            # new_obj["status"] = t_add
            try:
                if debug:
                    print(f'{new_obj["status"]} ==> {new_obj}')
                r = requests.post(
                    f"{n_env.api}/{t_job}", json=new_obj, headers=self.headers
                )
                r.raise_for_status()
                res = r.json()
                if res.get("status") == t_error or res.get("status") == t_skip:
                    raise ValueError(f"❌ Cannot add your file {obj.get('path')}")
                if debug:
                    print(f'{res["status"]} ==> {res}')
            except requests.exceptions.ConnectionError as err:
                print(error_busy, err)
                raise
            except requests.exceptions.HTTPError as err:
                print(error_reject, err)
                raise
            return new_obj
        else:
            raise ValueError(
                'obj should have all keys ("type","path","params","value")'
            )

    def del_prod(self, obj, debug):
        if "type" in obj and "path" in obj:
            new_obj = copy.copy(obj)
            new_obj["path"] = self.get_path(obj.get("path"))
            new_obj["params"] = {}
            new_obj["file"] = None
            new_obj["value"] = None
            new_obj["status"] = t_delete
            try:
                if debug:
                    print(f'{new_obj["status"]} ==> {new_obj}')
                r = requests.post(
                    f"{n_env.api}/{t_job}", json=new_obj, headers=self.headers
                )
                r.raise_for_status()
                res = r.json()
                if res.get("status") == t_error or res.get("status") == t_skip:
                    raise ValueError(f"❌ Cannot delete your file {obj.get('path')}")
                if debug:
                    print(f'{res.get("status")} ==> {res}')
            except requests.exceptions.ConnectionError as err:
                print(error_busy, err)
                raise
            except requests.exceptions.HTTPError as err:
                print(error_reject, err)
                raise
            return new_obj
        else:
            raise ValueError('obj should have keys ("type","path")')



from .env_var import n_env
import datetime as dt
import pandas as pd
import traceback
import json
import os

from .sqlite_table import SqliteTable


class Logger:
    __sql = None
    __name = "naas_logger"
    __logs_file = n_env.path_naas_folder + "/logs.db"
    __logs_csv_file = n_env.path_naas_folder + "/logs.csv"
    __date_format = "%Y-%m-%d %H:%M:%S.%f"
    __columns = [
        "asctime",
        "levelname",
        "name",
        "id",
        "type",
        "filename",
        "histo",
        "filepath",
        "output_filepath",
        "status",
        "error",
        "traceback",
        "duration",
        "url",
        "params",
        "token",
        "value",
        "main_id",
        "search",
    ]

    def __init__(self, clear=False):
        file_creation = not os.path.exists(self.__logs_file)
        #        is_csv = os.path.exists(self.__logs_csv_file)
        print("Init Naas logger")
        self.__sql = SqliteTable(self.__columns, self.__logs_file)
        if not file_creation and clear:
            self.__sql.clear()

    #        if file_creation and is_csv and not clear:
    #            self.__sql.csv_to_sql(self.__logs_csv_file)
    #            os.remove(self.__logs_csv_file)

    def add_log(self, **kwargs):
        kwargs["asctime"] = dt.datetime.now().strftime(self.__date_format)
        kwargs["name"] = self.__name
        return self.__sql.add_on_table(commit=True, **kwargs)

    def info(self, data):
        data["levelname"] = "INFO"
        return self.add_log(**data)

    def error(self, data):
        data["levelname"] = "ERROR"
        return self.add_log(**data)

    def clear(self):
        self.__sql.clear()

    def get_file_path(self):
        return self.__logs_file

    def list(
        self,
        uid: str,
        skip: int = 0,
        limit: int = 0,
        search: str = "",
        filters: list = [],
        sort: list = [],
        technical_rows: bool = True,
    ):
        df = None
        try:
            df = pd.DataFrame(
                data=self.__sql.search_in_db(search), index=None, columns=self.__columns
            )
            if not technical_rows and "type" in df and "filepath" in df:
                df = df[df["type"] != df["filepath"]]
                df = df[df["status"] != "busy"]
            if len(filters) > 0:
                df = df[df.type.isin(filters)]
            if len(sort) > 0:
                for query in sort:
                    field = [query["field"]]
                    ascending = False if query["type"] == "desc" else True
                    df = df.sort_values(by=field, ascending=ascending)
            total_records = len(df.index)
            if skip > 0:
                df = df.iloc[skip:]
            if limit > 0:
                df = df[:limit]
            df = df.reset_index()
            return {
                "uid": uid,
                "data": json.loads(df.to_json(orient="records")),
                "totalRecords": total_records,
            }
        except Exception as e:
            tb = traceback.format_exc()
            print("list logs", e, tb)
            return {"uid": uid, "data": [], "totalRecords": 0}



from naas.ntypes import t_tz, t_send, t_error, t_update
from sanic.views import HTTPMethodView
from naas.runner.env_var import n_env
from sanic import response
import uuid
import pytz


class TimezoneController(HTTPMethodView):
    __logger = None

    def __init__(self, logger, *args, **kwargs):
        super(TimezoneController, self).__init__(*args, **kwargs)
        self.__logger = logger

    def post(self, request):
        data = request.json
        tz = data.get("tz")
        uid = str(uuid.uuid4())
        if tz and tz in pytz.all_timezones:
            n_env.tz = tz
            self.__logger.info(
                {
                    "id": uid,
                    "type": t_tz,
                    "status": t_update,
                    "filepath": t_tz,
                }
            )
            return response.json({"tz": str(n_env.tz)})
        else:
            self.__logger.info(
                {
                    "id": uid,
                    "type": t_tz,
                    "status": t_error,
                    "filepath": t_tz,
                }
            )
            return response.json({"error": "this timezone don't exist"})

    async def get(self, request):
        uid = str(uuid.uuid4())
        self.__logger.info(
            {
                "id": uid,
                "type": t_tz,
                "status": t_send,
                "filepath": t_tz,
            }
        )
        return response.json({"tz": str(n_env.tz)})



from naas.ntypes import t_job, t_error, t_send, t_production
from sanic.views import HTTPMethodView
from naas.runner.env_var import n_env
from sanic import response
import base64
import errno
import uuid
import os
import datetime
import pytz

endpoint = "jobs"


class JobsController(HTTPMethodView):
    __jobs = None
    __logger = None
    __min_keys = sorted(list(["path", "type", "params", "value", "status", "file"]))

    def __init__(self, logger, jobs, *args, **kwargs):
        super(JobsController, self).__init__(*args, **kwargs)
        self.__jobs = jobs
        self.__logger = logger

    def __open_file(self, path, mode=None):
        filename = os.path.basename(path)
        if mode:
            filename = f"{mode}_{filename}"
        if not os.path.exists(path):
            data = bytes(f"File not found at path {path}", "utf-8")
            encoded = base64.b64encode(data)
            return {"filename": "Not_found.txt", "data": encoded.decode("ascii")}
        data = open(path, "rb").read()
        encoded = base64.b64encode(data)
        return {"filename": filename, "data": encoded.decode("ascii")}

    def __get_prod_path(self, path, filetype):
        # filename = os.path.basename(path)
        # dirname = os.path.dirname(path)
        # filename = f"{filetype}_{filename}"
        # path = os.path.join(dirname, filename)
        seps = os.sep + os.altsep if os.altsep else os.sep
        strip_path = os.path.splitdrive(path)[1].lstrip(seps)
        new_path = os.path.join(n_env.path_naas_folder, strip_path)
        return new_path

    def __save_file(self, path, data=None):
        if not data:
            raise FileNotFoundError(f"file doesn't have data {path}")
        if not os.path.exists(os.path.dirname(path)):
            try:
                os.makedirs(os.path.dirname(path))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        f = open(path, "wb")
        decoded = base64.b64decode(data)
        f.write(decoded)
        f.close()

    def __save_history(self, path, data=None):
        dt_string = datetime.datetime.now(tz=pytz.timezone(n_env.tz)).strftime(
            "%Y%m%d%H%M%S%f"
        )
        dirname = os.path.dirname(path)
        filename = os.path.basename(path)
        filename = f"{dt_string}___{filename}"
        new_path = os.path.join(dirname, filename)
        self.__save_file(new_path, data)

    async def put(self, request):
        uid = str(uuid.uuid4())
        reload_jobs = request.args.get("reload_jobs", False)
        move_job = request.args.get("move", False)
        cur_type = request.args.get("type", None)
        old_path = request.args.get("old_path", None)
        new_path = request.args.get("new_path", None)
        if reload_jobs:
            self.__jobs.reload_jobs()
            return response.json({"status": t_send})
        elif move_job and old_path and new_path:
            old_path = self.__get_prod_path(old_path, cur_type)
            new_path = self.__get_prod_path(new_path, cur_type)
            data = await self.__jobs.move_job(uid, old_path, new_path)
            return response.json(data)
        else:
            return response.json({"status": t_error})

    async def get(self, request):
        uid = str(uuid.uuid4())
        job = None
        cur_path = request.args.get("path", None)
        cur_type = request.args.get("type", None)
        cur_mode = request.args.get("mode", None)
        histo = request.args.get("histo", None)
        cur_light = request.args.get("light", False)
        if not cur_path or not cur_type:
            return response.json(await self.__jobs.list(uid))
        path = self.__get_prod_path(cur_path, cur_type)
        job = await self.__jobs.find_by_path(uid, path, cur_type)
        if not job:
            return response.json(
                {"status": t_error, "error": "job not found"}, status=404
            )
        if cur_mode and cur_mode == "list_history":
            job["files"] = self.__jobs.list_files(uid, path, cur_type)
        elif cur_mode and cur_mode == "list_output":
            job["files"] = self.__jobs.list_files(uid, path, cur_type, True)
        elif not cur_mode and not cur_light and not histo:
            job["file"] = self.__open_file(path, t_production)
        elif cur_mode or histo:
            new_path = "" + path
            dirname = os.path.dirname(new_path)
            filename = os.path.basename(new_path)
            if cur_mode and histo:
                filename = f"{histo}___{cur_mode}__{filename}"
            elif cur_mode:
                filename = f"{cur_mode}__{filename}"
            else:
                filename = f"{histo}___{filename}"
            new_path = os.path.join(dirname, filename)
            job["file"] = self.__open_file(new_path)
        self.__logger.info(
            {"id": uid, "type": t_job, "status": t_send, "filepath": endpoint}
        )
        return response.json(job)

    async def delete(self, request):
        uid = str(uuid.uuid4())
        histo = request.args.get("histo", None)
        cur_mode = request.args.get("mode", None)
        cur_path = request.args.get("path", None)
        cur_type = request.args.get("type", None)
        path = self.__get_prod_path(cur_path, cur_type)
        removed = self.__jobs.clear_file(uid, path, histo, cur_mode)
        return response.json(removed)

    async def post(self, request):
        uid = str(uuid.uuid4())
        data = request.json
        keys = sorted(list(data.keys()))
        if not data or self.__min_keys != keys:
            self.__logger.info(
                {
                    "id": uid,
                    "type": t_job,
                    "status": t_error,
                    "filepath": endpoint,
                    "error": "missing keys",
                    "tb": data,
                }
            )
            return response.json(
                {"id": uid, "status": t_error, "error": "missing keys", "data": [data]},
                status=400,
            )
        path = self.__get_prod_path(data["path"], data["type"])
        if data.get("file"):
            self.__save_file(path, data["file"]["data"])
            self.__save_history(path, data["file"]["data"])
        updated = await self.__jobs.update(
            uid,
            path,
            data["type"],
            data["value"],
            data["params"],
            data["status"],
        )
        if updated.get("error"):
            return response.json(updated, status=409)
        self.__logger.info(
            {
                "id": uid,
                "type": t_job,
                "filepath": endpoint,
                "status": updated["status"],
            }
        )
        return response.json(updated)
